#!/usr/bin/env python3

# Copyright (C) 2020  Braiins Systems s.r.o.
#
# This file is part of Braiins Open-Source Initiative (BOSI).
#
# BOSI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Please, keep in mind that we may also license BOSI or any part thereof
# under a proprietary license. For more information on the terms and conditions
# of such proprietary license or if you have any other questions, please
# contact us at opensource@braiins.com.

import argparse
import base64
import csv
import html.parser
import json
import logging
import os
import os.path
import re
import requests
import sys
import urllib.parse

from bos_toolbox.commands.boser.bos_gql_api import BosGQLApi
from bos_toolbox.platform.descriptor import TargetTriple, Target

LOG = logging.getLogger(__name__)

# these are in roughly same order as in web UI
fields = [
    'host',
    # format
    'model',
    # BOS management (deprecated)
    'bos_mgmt_id',
    # pool group
    'pool0_user',
    'pool0_password',
    'pool0_url',
    'pool1_user',
    'pool1_password',
    'pool1_url',
    'pool2_user',
    'pool2_password',
    'pool2_url',
    # hashchain global
    'asic_boost',
    'global_frequency',
    'global_voltage',
    # hashchain specifics
    'hashchainA_enabled',
    'hashchainA_frequency',
    'hashchainA_voltage',
    'hashchainB_enabled',
    'hashchainB_frequency',
    'hashchainB_voltage',
    'hashchainC_enabled',
    'hashchainC_frequency',
    'hashchainC_voltage',
    # thermal control
    'thermal_mode',
    'target_temp',
    'hot_temp',
    'dangerous_temp',
    # fanc control
    'fan_speed',
    'fan_min',
    # autotuning
    'autotuning',
    'power_target',
    # power scaling
    'power_scaling',
    'power_step',
    'min_power_target',
    'shutdown',
    'shutdown_duration',
    # non-bosminer options
    'auto_upgrade',
    # password is never loaded, but can be changed
    'password',
    # additional new fields
    'tuner_mode',
    'hashrate_target',
    'hashrate_step',
    'min_hashrate_target',
]

MINER_VERSION_REST = 1
MINER_VERSION_GRAPHQL = 2


def main(parser, args):
    logging.getLogger('gql.transport.aiohttp').disabled = True
    total = len(open(args.table, 'r').readlines())

    if args.action[0] in 'load':
        tmp_table = args.table + '.tmp'
        with open(args.table, newline='') as reader_fd, open(
            tmp_table, 'w', newline=''
        ) as writer_fd:
            reader = csv.DictReader(reader_fd, fields, restval='')
            writer = csv.DictWriter(
                writer_fd, fields, restval='', dialect=reader.dialect
            )
            writer.writerow({k: k for k in fields})

            for line_no, row in enumerate(reader, 1):
                host = row['host'].strip()
                out = {'host': host}
                try:
                    if (
                        line_no == 1 and host == fields[0] or host == ''
                    ):  # skip first line with headers
                        continue

                    version = detect_miner_version(host)
                    LOG.info(
                        'Pulling from %s (%d/%d) version=%d'
                        % (host, line_no, total, version)
                    )

                    if version == MINER_VERSION_REST:
                        api = BosApi(host, args.user, args.password)
                        out['auto_upgrade'] = toggle2str(
                            api.uci_get_bool_legacy('bos', 'auto_upgrade', 'enable')
                        )
                        cfg, triple = api.get_config()
                        v2 = False
                    elif version == MINER_VERSION_GRAPHQL:
                        api = BosGQLApi()
                        api.connect_and_authenticate(host, args.user, args.password)
                        cfg, triple = api.get_config()
                        v2 = api.v2
                        out['auto_upgrade'] = toggle2str(cfg['auto_upgrade'])

                    # output row data
                    csvizer = Csvizer(cfg, triple, v2)
                    csvizer.pull(out)
                    writer.writerow(out)
                except Exception as ex:
                    log_error(row)
                    LOG.warning(str(ex))
                    writer.writerow(out)

        if not args.check:
            # only replace original file once we are all done so as not to clobber it if error occurs
            os.replace(args.table, args.table + '.bak')
            os.replace(tmp_table, args.table)
        else:
            os.unlink(tmp_table)

    elif args.action[0] in ('save', 'save_apply'):
        with open(args.table, newline='') as reader_fd:
            reader = csv.DictReader(reader_fd, fields, restval='')

            for line_no, row in enumerate(reader, 1):
                try:
                    host = row['host'].strip()
                    password = row['password'].strip()
                    if line_no == 1 and host == fields[0]:
                        # if column headers are present, make sure they fit
                        for name in fields:
                            if name != row[name]:
                                sys.exit(
                                    'Mismatched header: %s (expected %s)'
                                    % (row[name], name)
                                )
                        continue  # skip header row
                    if host == '':
                        continue  # skip over empty hosts
                    auto_upgrade = str2toggle(row['auto_upgrade'])
                    LOG.info('Pushing to %s (%d/%d)' % (host, line_no, total))

                    version = detect_miner_version(host)
                    api = (
                        BosApi(host, args.user, args.password)
                        if version == MINER_VERSION_REST
                        else BosGQLApi().connect_and_authenticate(
                            host, args.user, args.password
                        )
                    )
                    cfg, triple = api.get_config()
                    v2 = api.v2 if version == MINER_VERSION_GRAPHQL else False
                    csvizer = Csvizer(cfg, triple, v2)
                    csvizer.push(row)
                    if version == MINER_VERSION_GRAPHQL:
                        csvizer.cfg['auto_upgrade'] = str2toggle(row['auto_upgrade'])
                    if not args.check:
                        api.set_config(csvizer.cfg, args.action[0] in 'save_apply')
                        if auto_upgrade is not None:
                            api.uci_init_section('bos', 'auto_upgrade', 'bos')
                            ret = api.uci_set_bool(
                                'bos', 'auto_upgrade', 'enable', auto_upgrade
                            )
                            api.uci_commit('bos')
                        if args.action[0] in 'save_apply':
                            api.apply_config()
                        # password is part of system administration, not miner config
                        if password and args.change_password:
                            LOG.info(
                                'Changing password for %s (%d/%d)'
                                % (host, line_no, total)
                            )
                            api.set_password(password)
                except Exception as ex:
                    log_error(row)
                    LOG.warning(str(ex))

    elif args.action[0] in 'apply':
        with open(args.table, newline='') as reader_fd:
            reader = csv.DictReader(reader_fd, fields, restval='')
            for line_no, row in enumerate(reader, 1):
                try:
                    host = row['host'].strip()
                    if (
                        line_no == 1 and host == fields[0] or host == ''
                    ):  # skip first line with headers
                        continue
                    LOG.info('Applying on %s (%d/%d)' % (host, line_no, total))
                    api = BosApi(host, args.user, args.password)
                    if not args.check:
                        api.apply_config()
                except Exception as ex:
                    log_error(row)
                    LOG.warning(str(ex))

    else:
        raise RuntimeError('unknown action')


def detect_miner_version(host):
    response = requests.api.get(f'http://{host}/graphql', timeout=5)
    return MINER_VERSION_REST if response.status_code == 404 else MINER_VERSION_GRAPHQL


class Csvizer:
    GROUP_SIZE = 3  # how many pools to have in our group
    DEFAULT_GROUP = {'name': 'default', 'pool': []}  # defaults for created pool group

    def __init__(self, cfg, triple, v2):
        cfg_format = cfg['format']
        cfg_format_version = cfg_format['version']
        cfg_format_model = cfg_format['model']

        # json struct with config to work on
        self.cfg = cfg
        self._cfg_format_version = cfg_format_version
        self._cfg_format_model = cfg_format_model
        self._triple = triple
        self.v2 = v2

    def is_bos_plus(self) -> bool:
        return (
            self._cfg_format_version['isPlus']
            if isinstance(self._cfg_format_version, dict)
            else self._cfg_format_version.endswith('+')
        )

    def is_antminer_s9(self) -> bool:
        return self._triple.target_full_name == 'am1-s9'

    def is_antminer_x17(self) -> bool:
        return self._triple.target_full_name == 'am2-x17'

    def is_whatsminer(self) -> bool:
        return self._triple.target == Target.wm1

    def _get_first_chain_index(self):
        return 1 if self.is_antminer_x17() or self.is_whatsminer() else 6

    def _has_chain_voltage(self):
        return self.is_antminer_s9()

    def pull(self, row):
        """
        Gather data from existing config into dict suitable for csvwriter.
        row: dict to be populated with values pulled from config
        """
        row['model'] = self._cfg_format_model
        self.pull_groups(row)
        self.pull_hashchain_global(row)
        self.pull_fanctl(row)
        self.pull_tempctl(row)
        # hashchains are currently hardcoded in bosminer
        for chain in range(ord('A'), ord('C') + 1):
            self.pull_hashchain(row, chr(chain))
        # bosminer plus additional features
        if self.is_bos_plus():
            self.pull_power_scaling(row)
            self.pull_autotuning(row)

    def push(self, row):
        """
        Use data in dict to update existing config.
        row: dict to be populated with values pulled from config
        """
        # options from plus will bug out regular bosminer
        if self.is_bos_plus():
            self.push_power_scaling(row)
            self.push_autotuning(row)

        self.push_groups(row)
        self.push_hashchain_global(row)
        self.push_fanctl(row)
        self.push_tempctl(row)
        for chain in range(ord('A'), ord('C') + 1):
            self.push_hashchain(row, chr(chain))

    # pools --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
    def pull_groups(self, row):
        """
        row: dict to be populated with values pulled from config
        for now, only four pools in first group are considered, rest is ignored
        """
        if 'group' in self.cfg and len(self.cfg['group']) > 0:
            # gather the first group, rest is ignored
            for i, pool in enumerate(
                self.cfg['group'][0].get('pool', [])[: self.GROUP_SIZE]
            ):
                row['pool%d_url' % i] = pool.get('url')
                row['pool%d_user' % i] = pool.get('user')
                row['pool%d_password' % i] = pool.get('password')

    def push_groups(self, row):
        """
        row: dict to pull new values from
        no fancy merging yet, if there is anything in four pools in csv, first group is overwriten with that
        """
        if 'group' not in self.cfg:
            self.cfg['group'] = [{'name': 'default', 'pool': []}]
        pools = []
        for i in range(self.GROUP_SIZE):
            pool = {
                'enabled': True,
                'user': row.get('pool%d_user' % i, ''),
                'password': row.get('pool%d_password' % i, ''),
                'url': str2url(row.get('pool%d_url' % i, '')),
            }
            if pool['url'] and pool['user']:
                pools.append(pool)
            elif pool['url'] and not pool['user']:
                raise InvalidUser('(unspecified)')
            elif not pool['url'] and pool['user']:
                raise InvalidUrl('(unspecified)')
            else:
                pass  # neither url nor user == just ignore this
        if pools:
            self.cfg.setdefault('group', [self.DEFAULT_GROUP])[0]['pool'] = pools

    # autotuning --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
    def pull_autotuning(self, row):
        """
        row: dict to be populated with values pulled from config
        Otherwise logic is reverse of push.
        """
        row['autotuning'] = toggle2str(self.cfg.get('autotuning', {}).get('enabled'))
        row['power_target'] = self.cfg.get('autotuning', {}).get('psu_power_limit', '')
        row['hashrate_target'] = self.cfg.get('autotuning', {}).get(
            'hashrate_target', ''
        )
        row['tuner_mode'] = self.cfg.get('autotuning', {}).get('mode', '')

    def push_autotuning(self, row):
        """
        row: dict to pull new values from
        """
        enabled = str2toggle(row['autotuning'])
        power_target = str2int(row['power_target'], allow_empty=True)

        if enabled is not None:
            self.cfg.setdefault('autotuning', {})['enabled'] = enabled
        if power_target is not None:
            self.cfg.setdefault('autotuning', {})['psu_power_limit'] = power_target

        if self.v2:
            tuner_mode = row['tuner_mode']
            hashrate_target = str2float(row['hashrate_target'], allow_empty=True)

            if tuner_mode is not None:
                self.cfg.setdefault('autotuning', {})['mode'] = tuner_mode
            if hashrate_target is not None:
                self.cfg.setdefault('autotuning', {})[
                    'hashrate_target'
                ] = hashrate_target
        else:
            tuner_mode = row['tuner_mode']
            if tuner_mode not in ['POWER_TARGET', '', None]:
                raise UnsupportedTunerMode(tuner_mode)

    # fan control --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
    def pull_fanctl(self, row):
        """
        row: dict to be populated with values pulled from config
        """
        row['fan_speed'] = self.cfg.get('fan_control', {}).get('speed', '')
        row['fan_min'] = self.cfg.get('fan_control', {}).get('min_fans', '')

    def push_fanctl(self, row):
        """
        row: dict to pull new values from
        TODO: how to handle empty fields?
        """

        speed = str2int(row['fan_speed'], allow_empty=True)
        if speed is not None:
            self.cfg.setdefault('fan_control', {})['speed'] = speed

        min_fans = str2int(row['fan_min'], allow_empty=True)
        if min_fans is not None:
            if min_fans == 0:  # user want to turn on immersion
                immersion_is_off_on_miner = (
                    self.cfg.get('fan_control', {})['min_fans'] > 0
                )
                self.cfg['fan_control'] = {}
                if immersion_is_off_on_miner:  # Immersion is turned off on miner
                    self.cfg['fan_control']['immersion_mode_enabled'] = True
            elif min_fans > 0:
                if self.cfg.get('fan_control', {})['min_fans'] == 0:
                    # User wants to disable immersion
                    self.cfg['fan_control']['immersion_mode_enabled'] = False
                self.cfg.setdefault('fan_control', {})['min_fans'] = min_fans
            else:
                raise Exception('min_fans must be 0(immersion) or higher')
        if (
            min_fans is None
            and 'min_fans' in self.cfg.get('fan_control', {})
            and self.cfg.get('fan_control', {})['min_fans'] == 0
        ):
            # To not send anything control if immersion is turned on and w are not editing min_fans
            self.cfg['fan_control'] = {}

    # thermal  --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
    def pull_tempctl(self, row):
        """
        row: dict to be populated with values pulled from config
        """
        row['thermal_mode'] = self.cfg.get('temp_control', {}).get('mode', '')
        row['target_temp'] = self.cfg.get('temp_control', {}).get('target_temp', '')
        row['hot_temp'] = self.cfg.get('temp_control', {}).get('hot_temp', '')
        row['dangerous_temp'] = self.cfg.get('temp_control', {}).get(
            'dangerous_temp', ''
        )

    def push_tempctl(self, row):
        """
        row: dict to pull new values from
        TODO: how to handle empty fields again?
        """
        mode = row['thermal_mode'].strip().lower()
        if mode not in ('auto', 'manual', 'disabled', ''):
            raise InvalidThermalMode(mode)
        target_temp = str2float(row['target_temp'], allow_empty=True)
        hot_temp = str2float(row['hot_temp'], allow_empty=True)
        dangerous_temp = str2float(row['dangerous_temp'], allow_empty=True)

        if mode:
            self.cfg.setdefault('temp_control', {})['mode'] = mode
        if target_temp is not None:
            self.cfg.setdefault('temp_control', {})['target_temp'] = target_temp
        if hot_temp is not None:
            self.cfg.setdefault('temp_control', {})['hot_temp'] = hot_temp
        if dangerous_temp is not None:
            self.cfg.setdefault('temp_control', {})['dangerous_temp'] = dangerous_temp

    # global hashchain  --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
    def pull_hashchain_global(self, row):
        """
        row: dict to be populated with values pulled from config
        """
        row['asic_boost'] = toggle2str(
            self.cfg.get('hash_chain_global', {}).get('asic_boost')
        )
        row['global_frequency'] = self.cfg.get('hash_chain_global', {}).get(
            'frequency', ''
        )
        row['global_voltage'] = self.cfg.get('hash_chain_global', {}).get('voltage', '')

    def push_hashchain_global(self, row):
        """
        row: dict to pull new values from
        TODO: how to handle empty fields again?
        """
        asic_boost = str2toggle(row['asic_boost'])
        frequency = str2float(row['global_frequency'], allow_empty=True)
        voltage = str2float(row['global_voltage'], allow_empty=True)

        if asic_boost is not None:
            self.cfg.setdefault('hash_chain_global', {})['asic_boost'] = asic_boost
        if frequency is not None:
            self.cfg.setdefault('hash_chain_global', {})['frequency'] = frequency
        if voltage is not None:
            self.cfg.setdefault('hash_chain_global', {})['voltage'] = voltage

    # specific hashchain  --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
    def get_chain_index(self, chain) -> str:
        return str(self._get_first_chain_index() + ord(chain) - ord('A'))

    def pull_hashchain(self, row, chain):
        """
        row: dict to be populated with values pulled from config
        chain: chain number (6-8)
        """
        chain_index = self.get_chain_index(chain)
        hash_chain = self.cfg.get('hash_chain', {})
        hash_chain = (
            hash_chain.get(chain_index, {})
            if isinstance(hash_chain, dict)
            else next(
                (chain for chain in hash_chain if chain['name'] == chain_index), {}
            )
        )
        row['hashchain{}_enabled'.format(chain)] = toggle2str(hash_chain.get('enabled'))
        row['hashchain{}_frequency'.format(chain)] = hash_chain.get('frequency', '')
        row['hashchain{}_voltage'.format(chain)] = hash_chain.get('voltage', '')

    def push_hashchain(self, row, chain):
        """
        row: dict to pull new values from
        chain: chain number (6-8)
        """
        autotuning = self.cfg.get('autotuning', {}).get('enabled', False)
        chain_index = self.get_chain_index(chain)
        enabled = str2toggle(row['hashchain{}_enabled'.format(chain)])
        frequency = str2float(
            row['hashchain{}_frequency'.format(chain)], allow_empty=True
        )
        voltage = str2float(row['hashchain{}_voltage'.format(chain)], allow_empty=True)
        if not self._has_chain_voltage() and voltage is not None:
            LOG.warning(
                "Attribute 'voltage' is ignored for {}".format(self._cfg_format_model)
            )
            voltage = None
        if autotuning and voltage is not None:
            LOG.warning("Attribute 'voltage' is ignored for auto-tuning")
            voltage = None
        if autotuning and frequency is not None:
            LOG.warning("Attribute 'frequency' is ignored for auto-tuning")
            frequency = None

        if isinstance(self.cfg.get('hash_chain'), list):
            self.cfg['hash_chain'] = {x['name']: x for x in self.cfg['hash_chain']}

        if enabled is not None:
            self.cfg.setdefault('hash_chain', {}).setdefault(chain_index, {})[
                'enabled'
            ] = enabled
        if frequency is not None:
            self.cfg.setdefault('hash_chain', {}).setdefault(chain_index, {})[
                'frequency'
            ] = frequency
        if voltage is not None:
            self.cfg.setdefault('hash_chain', {}).setdefault(chain_index, {})[
                'voltage'
            ] = voltage

    # power scaling  --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
    def pull_power_scaling(self, row):
        """
        row: dict to be populated with values pulled from config
        """
        row['power_scaling'] = toggle2str(
            self.cfg.get('power_scaling', {}).get('enabled')
        )
        row['min_power_target'] = self.cfg.get('power_scaling', {}).get(
            'min_psu_power_limit'
        )
        row['power_step'] = self.cfg.get('power_scaling', {}).get('power_step')
        row['min_hashrate_target'] = self.cfg.get('power_scaling', {}).get(
            'min_hashrate_target'
        )
        row['hashrate_step'] = self.cfg.get('power_scaling', {}).get('hashrate_step')
        row['shutdown'] = toggle2str(
            self.cfg.get('power_scaling', {}).get('shutdown_enabled')
        )
        row['shutdown_duration'] = self.cfg.get('power_scaling', {}).get(
            'shutdown_duration'
        )

    def push_power_scaling(self, row):
        enabled = str2toggle(row['power_scaling'])
        min_power = str2int(row['min_power_target'], allow_empty=True)
        power_step = str2int(row['power_step'], allow_empty=True)
        shutdown = str2toggle(row['shutdown'])
        duration = str2float(row['shutdown_duration'], allow_empty=True)

        if enabled is not None:
            self.cfg.setdefault('power_scaling', {})['enabled'] = enabled
        if min_power is not None:
            self.cfg.setdefault('power_scaling', {})['min_psu_power_limit'] = min_power
        if power_step is not None:
            self.cfg.setdefault('power_scaling', {})['power_step'] = power_step
        if shutdown is not None:
            self.cfg.setdefault('power_scaling', {})['shutdown_enabled'] = shutdown
        if duration is not None:
            self.cfg.setdefault('power_scaling', {})['shutdown_duration'] = duration

        if self.v2:
            min_hashrate = str2float(row['min_hashrate_target'], allow_empty=True)
            hashrate_step = str2float(row['hashrate_step'], allow_empty=True)

            if min_hashrate is not None:
                self.cfg.setdefault('power_scaling', {})[
                    'min_hashrate_target'
                ] = min_hashrate
            if hashrate_step is not None:
                self.cfg.setdefault('power_scaling', {})[
                    'hashrate_step'
                ] = hashrate_step


class CsvizerError(RuntimeError):
    def __str__(self):
        return '%s: %s' % (self.hint, self.args[0])


class InvalidNumber(CsvizerError):
    hint = 'invalid number'


class InvalidThermalMode(CsvizerError):
    hint = 'invalid thermal mode'


class InvalidToggle(CsvizerError):
    hint = 'invalid toggle'


class InvalidUrl(CsvizerError):
    hint = 'invalid pool url'


class InvalidUser(CsvizerError):
    hint = 'invalid pool user'


class UnsupportedTunerMode(CsvizerError):
    hint = 'unsupported tuner mode'


class LuciAdminFormParser(html.parser.HTMLParser):
    """
    Helper for getting csrf token in administration web
    """

    def __init__(self):
        super().__init__()
        self.token = None

    def handle_starttag(self, tag, attrs):
        attrs = dict(attrs)
        if (
            tag == 'input'
            and attrs.get('type') == 'hidden'
            and attrs.get('name') == 'token'
        ):
            self.token = attrs.get('value')


class BosApi:
    """
    Wrapper over bosminer json api. It's main function is to get and hold authentication cookie.
    """

    timeout = 3

    def __init__(self, host, username='root', password=''):
        self.host = host
        self.username = username
        self.password = password
        self.auth()
        self.rpc_id_counter = 0
        self.supported_versions = {
            'am1-s9': ((1, 0), (1, 2)),
            'am2-x17': ((1, 0), (1, 2)),
        }

    def check_config(self, data) -> TargetTriple:
        """Check if retrieved config is in known format"""
        try:
            format = data['format']
            model = format['model']
            triple = TargetTriple._get_target_triple(model)
            if not triple:
                raise UnsupportedDevice(model)

            version = format['version']
            supported_versions = self.supported_versions[triple.target_full_name]

            if version.endswith('+'):
                version = version[:-1]
                ver_idx = 1
            else:
                ver_idx = 0

            ver_major, ver_minor = map(lambda v: int(v), version.split('.'))
            ver_range = supported_versions[ver_idx]

            if not (1 <= ver_major <= ver_range[0] and 0 <= ver_minor <= ver_range[1]):
                raise UnsupportedVersion(version)
        except KeyError as ex:
            raise ConfigStructError(str(ex))
        else:
            return triple

    def get_url(self, path):
        """Construct api url from known hostname and given path"""
        url = urllib.parse.urlunsplit(
            ('http', self.host, path, '', '')  # query  # fragment
        )
        return url

    def auth(self):
        """Acquire an authentication token"""
        response = requests.post(
            self.get_url('/cgi-bin/luci/'),
            data={'luci_username': self.username, 'luci_password': self.password},
            allow_redirects=False,
            timeout=self.timeout,
        )
        response.raise_for_status()
        self.authcookie = response.cookies.copy()

    def get_config(self):
        """Load miner config"""
        response = requests.get(
            self.get_url('/cgi-bin/luci/admin/miner/cfg_data'),
            cookies=self.authcookie,
            timeout=self.timeout,
        )
        response.raise_for_status()
        document = response.json()
        if document.get('status', {}).get('code') != 0:
            raise ConfigFetchFailed(document)
        triple = self.check_config(document.get('data'))
        return document['data'], triple

    def set_config(self, data, _apply):
        """Write miner config"""
        data['format']['generator'] = 'multiconfiger 0.1'  #
        response = requests.post(
            self.get_url('/cgi-bin/luci/admin/miner/cfg_save'),
            cookies=self.authcookie,
            data=json.dumps({'data': data}),
            timeout=self.timeout,
        )
        response.raise_for_status()
        document = response.json()
        if document.get('status', {}).get('code') != 0:
            raise ConfigStoreFailed(document)
        # print(response.content)

    def apply_config(self):
        """Restart bosminer"""
        response = requests.post(
            self.get_url('/cgi-bin/luci/admin/miner/cfg_apply'),
            cookies=self.authcookie,
            allow_redirects=False,
            timeout=self.timeout,
        )
        response.raise_for_status()
        document = response.json()
        if document.get('status', {}).get('code') != 0:
            raise ConfigApplyFailed(document)

    def rpc(self, endpoint, method, *args):
        """Perform a rpc call"""
        # see https://github.com/openwrt/luci/wiki/JsonRpcHowTo
        rpc = {'method': method, 'params': args, 'id': self.rpc_id()}
        response = requests.post(
            self.get_url('/cgi-bin/luci/rpc/%s' % endpoint),
            cookies=self.authcookie,
            allow_redirects=False,
            data=json.dumps(rpc),
        )
        if response.status_code == 404:
            raise UnsupportedVersion('missing rpc interface')
        response.raise_for_status()
        document = response.json()
        if document.get('id') != rpc['id']:
            raise RpcError('mismatched rpc call (%s)' % document.get('id'))
        if document.get('error'):
            raise RpcError(document.get('error'))
        return document.get('result')

    def rpc_id(self):
        """Provide unique id required by json-rpc spec"""
        self.rpc_id_counter += 1
        return self.rpc_id_counter

    def uci_commit(self, config):
        return self.rpc('uci', 'commit', config)

    def uci_init_section(self, config, section, type):
        """Create section if it does not exist"""
        if self.rpc('uci', 'get', config, section) is None:
            self.rpc('uci', 'set', config, section, type)

    def uci_get(self, config, section, option):
        return self.rpc('uci', 'get', config, section, option)

    def uci_set(self, config, section, option, value):
        return self.rpc('uci', 'set', config, section, option, value)

    def uci_get_bool(self, config, section, option):
        value = self.uci_get(config, section, option)
        # may get nil if option or section is not present
        if value not in ('0', '1', None):
            raise RpcError('Invalid bool value: %s' % value)
        return {'0': False, '1': True}.get(value)

    def uci_get_bool_legacy(self, config, section, option):
        try:
            return self.uci_get_bool(config, section, option)
        except UnsupportedVersion as ex:
            return None

    def uci_set_bool(self, config, section, option, value):
        return self.uci_set(config, section, option, '1' if value else '0')

    def fs_readfile(self, path):
        result = self.rpc('fs', 'readfile', path)
        return result and base64.b64decode(result)

    def fs_writefile(self, path, data: bytes):
        data = base64.b64encode(data).decode()
        return self.rpc('fs', 'writefile', path, data) == 1

    def set_password(self, password):

        # we need admin page to obtain csrf token
        response = requests.get(
            self.get_url('/cgi-bin/luci/admin/system/admin'), cookies=self.authcookie
        )
        response.raise_for_status()

        parser = LuciAdminFormParser()
        parser.feed(response.content.decode('utf8'))
        if not parser.token:
            raise MissingToken()

        # save the password. there is more in the form which we skip over
        # using None for filename is undocumented feature of requests lib
        form = {
            'token': (None, parser.token),
            'cbi.submit': (None, '1'),
            'cbid.system._pass.pw1': (None, password),
            'cbid.system._pass.pw2': (None, password),
        }
        response = requests.post(
            self.get_url('/cgi-bin/luci/admin/system/admin'),
            cookies=self.authcookie,
            allow_redirects=False,
            files=form,
        )
        response.raise_for_status()


class BosApiError(RuntimeError):
    pass


class ErrorResponse(BosApiError):
    def __init__(self, document):
        super().__init__(
            '%s (%s)'
            % (
                document.get('status', {}).get('code'),
                document.get('status', {}).get('message'),
            )
        )


class MissingToken(BosApiError):
    pass


class UnsupportedDevice(BosApiError):
    pass


class UnsupportedVersion(BosApiError):
    pass


class ConfigStructError(BosApiError):
    pass


class ConfigFetchFailed(ErrorResponse):
    hint = 'configuration load failed'


class ConfigStoreFailed(ErrorResponse):
    hint = 'configuration save failed'


class ConfigApplyFailed(ErrorResponse):
    hint = 'miner state change failed'


class RpcError(BosApiError):
    pass


# --- helper functions --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
def str2int(s, exception=InvalidNumber, message=None, allow_empty=False):
    """
    Helper function to raise an error if conversion from string to int fails
    """
    if allow_empty and s.strip() == '':
        return None
    try:
        return int(s)
    except ValueError:
        raise exception(message or s)


def str2float(s, exception=InvalidNumber, message=None, allow_empty=False):
    """
    Helper function to raise an error if conversion from string to float fails
    """
    if allow_empty and s.strip() == '':
        return None
    try:
        return float(s)
    except ValueError:
        raise exception(message or s)


def str2toggle(b):
    """
    Turn 'enabled' or 'disabled' strings into bool, or empty string into None
    Exception is thrown for other values.
    """
    b = b.strip().lower()
    if b == '':
        return None
    elif b == 'disabled':
        return False
    elif b == 'enabled':
        return True
    raise InvalidToggle(b)


def toggle2str(b):
    """
    Turn bool config value into enabled/disabled string.
    None is returned as empty string, which should represent value not present.
    (which is why this is not called bool2str)
    """
    return {None: '', False: 'disabled', True: 'enabled'}[b]


def str2url(u):
    """
    Turns string unto url, which is string again so ti does not do anything really.
    But it checks validity, so we can throw more reasonable error then what server does.
    """
    # TODO: this can be dropped once server error messages get better then rust panics
    if u == '':
        return None

    if re.match(
        r'(?:drain|(?:stratum2?\+tcp(?:\+insecure)?)):\/\/[\w\.-]+(?::\d+)?(?:\/[\dA-HJ-NP-Za-km-z]+)?',
        u,
    ):
        return u

    raise InvalidUrl(u)

    if u.strip() == '':
        return None
    parsed = urllib.parse.urlparse(u)
    if parsed.scheme not in ('stratum+tcp', 'stratum2+tcp'):
        raise InvalidUrl(u)
    if parsed.netloc == '':
        raise InvalidUrl(u)


def log_error(data=None):
    import traceback
    import time

    with open('error.log', 'a') as fd:
        print(
            time.strftime('%y-%m-%d %H:%M:%S')
            + (' '.join(sys.argv) if log_error.count == 0 else ''),
            file=fd,
        )
        if data:
            print(data, file=fd)
        traceback.print_exc(file=fd)


log_error.count = 0


def build_arg_parser(parser, desc):
    label = desc['label']
    hlp = desc['help']

    parser.description = desc['description']
    parser.add_argument(
        'action',
        nargs='*',
        type=lambda x: x.strip().lower(),
        metavar=label['action'],
        help=hlp['action'],
        choices=['load', 'save', 'apply', 'save_apply'],
        # widget='FilterableDropdown',
        widget='Listbox',
        gooey_options={
            # height of the listbox in pixels
            'height': 105,
            'show_help': False,
        },
    )
    parser.add_argument(
        'table', metavar=label['table'], help=hlp['table'], widget='FileChooser'
    )
    parser.add_argument(
        '-u', '--user', default='root', metavar=label['user'], help=hlp['user']
    )
    parser.add_argument(
        '-p',
        '--password',
        default='',
        metavar=label['password'],
        help=hlp['password'],
        widget='PasswordField',
    )
    parser.add_argument(
        '--change-password',
        action='store_true',
        metavar=label['change-password'],
        help=hlp['change-password'],
    )
    parser.add_argument(
        '-c',
        '--check',
        action='store_true',
        metavar=label['check'],
        help=hlp['check'],
        gooey_options={'visible': False},
    )


if __name__ == '__main__':
    try:
        parser = argparse.ArgumentParser()
        build_arg_parser(parser)
        args = parser.parse_args()
        main(parser, args)
    except ErrorResponse as ex:
        # error messages inside responses can be very ugly with rust panic strings
        # handle them with special care until they get better
        log_error()
        sys.exit('error: %s' % ex.hint)
    except Exception as ex:
        log_error()
        sys.exit('error: %s' % ex)
